' Copyright 2010 Microsoft Corporation
' Licensed under the Apache License, Version 2.0 (the "License")
' You may not use this file except in compliance with the License.
' You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0

' THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR
' CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED,
' INCLUDING WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR
' CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
' MERCHANTABLITY OR NON-INFRINGEMENT.

' See the Apache 2 License for the specific language governing
' permissions and limitations under the License.

Imports System.Text
Imports System.Security.Cryptography

Namespace Microsoft.AccessControl2.SDK
	Public Class TokenValidator
		Private issuerLabel As String = "Issuer"
		Private expiresLabel As String = "ExpiresOn"
		Private audienceLabel As String = "Audience"
		Private hmacSHA256Label As String = "HMACSHA256"

		Private acsHostName As String

		Private trustedSigningKey As String
		Private trustedTokenIssuer As String
		Private trustedAudienceValue As String

		Public Sub New(ByVal acsHostName As String, ByVal serviceNamespace As String, ByVal trustedAudienceValue As String, ByVal trustedSigningKey As String)
			Me.trustedSigningKey = trustedSigningKey
			Me.trustedTokenIssuer = String.Format("https://{0}.{1}/", serviceNamespace.ToLowerInvariant, acsHostName.ToLowerInvariant)

			Me.trustedAudienceValue = trustedAudienceValue
		End Sub

		Public Function Validate(ByVal token As String) As Boolean
			If Not Me.IsHMACValid(token, Convert.FromBase64String(Me.trustedSigningKey)) Then
				Return False
			End If
			Trace.WriteLine("IsHMACValid")

			If Me.IsExpired(token) Then
				Return False
			End If
			Trace.WriteLine("IsExpired")

			If Not Me.IsIssuerTrusted(token) Then
				Return False
			End If
			Trace.WriteLine("IsIssuerTrusted")

			If Not Me.IsAudienceTrusted(token) Then
				Return False
			End If
			Trace.WriteLine("IsAudienceTrusted")

			Return True
		End Function

		Public Function GetNameValues(ByVal token As String) As Dictionary(Of String, String)
			If String.IsNullOrEmpty(token) Then
				Throw New ArgumentException
			End If

			Return token.Split("&"c).Aggregate(New Dictionary(Of String, String), Function(dict, rawNameValue)
				If rawNameValue = String.Empty Then
					Return dict
				End If
				Dim nameValue() As String = rawNameValue.Split("="c)
				If nameValue.Length <> 2 Then
					Throw New ArgumentException("Invalid formEncodedstring - contains a name/value pair missing an = character")
				End If
				If dict.ContainsKey(nameValue(0)) = True Then
					Throw New ArgumentException("Repeated name/value pair in form")
				End If
				dict.Add(HttpUtility.UrlDecode(nameValue(0)), HttpUtility.UrlDecode(nameValue(1)))
				Return dict
			End Function)
		End Function

		Private Shared Function GenerateTimeStamp() As ULong
			' Default implementation of epoch time
			Dim ts As TimeSpan = Date.UtcNow - New Date(1970, 1, 1, 0, 0, 0, 0)
			Return Convert.ToUInt64(ts.TotalSeconds)
		End Function

		Private Function IsAudienceTrusted(ByVal token As String) As Boolean
			Dim tokenValues As Dictionary(Of String, String) = Me.GetNameValues(token)

            Dim audienceValue As String = String.Empty

			tokenValues.TryGetValue(Me.audienceLabel, audienceValue)

			If Not String.IsNullOrEmpty(audienceValue) Then
				If audienceValue.Equals(Me.trustedAudienceValue, StringComparison.Ordinal) Then
					Return True
				ElseIf HttpUtility.UrlDecode(audienceValue).Equals(Me.trustedAudienceValue, StringComparison.Ordinal) Then
					Return True
				End If
			End If

			Return False
		End Function

		Private Function IsIssuerTrusted(ByVal token As String) As Boolean
			Dim tokenValues As Dictionary(Of String, String) = Me.GetNameValues(token)

            Dim issuerName As String = String.Empty

			tokenValues.TryGetValue(Me.issuerLabel, issuerName)

			If Not String.IsNullOrEmpty(issuerName) Then
				If issuerName.Equals(Me.trustedTokenIssuer) Then
					Return True
				End If
			End If

			Return False
		End Function

		Private Function IsHMACValid(ByVal swt As String, ByVal sha256HMACKey() As Byte) As Boolean
			Dim swtWithSignature() As String = swt.Split(New String() { "&" & Me.hmacSHA256Label & "=" }, StringSplitOptions.None)

			If (swtWithSignature Is Nothing) OrElse (swtWithSignature.Length <> 2) Then
				Return False
			End If

			Dim hmac As New HMACSHA256(sha256HMACKey)

			Dim locallyGeneratedSignatureInBytes() As Byte = hmac.ComputeHash(Encoding.ASCII.GetBytes(swtWithSignature(0)))

			Dim locallyGeneratedSignature As String = HttpUtility.UrlEncode(Convert.ToBase64String(locallyGeneratedSignatureInBytes))

			Return locallyGeneratedSignature = swtWithSignature(1)
		End Function

		Private Function IsExpired(ByVal swt As String) As Boolean
			Try
				Dim nameValues As Dictionary(Of String, String) = Me.GetNameValues(swt)
				Dim expiresOnValue As String = nameValues(Me.expiresLabel)
				Dim expiresOn As ULong = Convert.ToUInt64(expiresOnValue)
				Dim currentTime As ULong = Convert.ToUInt64(GenerateTimeStamp)

				If currentTime > expiresOn Then
					Return True
				End If

				Return False
			Catch e1 As KeyNotFoundException
				Throw New ArgumentException
			End Try
		End Function
	End Class
End Namespace
